/**
 * @file test_dummy_coupling4_c.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "tests.h"
#include "test_common.h"
#include "yac_interface.h"

int main (void) {

  yac_cinit("coupling_test4.xml", "coupling.xsd");

  int size, rank;
  MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
  MPI_Comm_size ( MPI_COMM_WORLD, &size );

  if (size != 2) {
    fputs("wrong number of processes (has to be 2)\n", stderr);
    exit(EXIT_FAILURE);
  }

  int is_target = rank == 1;

  // define local component
  int comp_id;
  yac_cdef_comp((is_target)?"target_comp":"source_comp", &comp_id);

  // define grid (both components use an identical grid
  int grid_id;
  yac_cdef_grid_reg2d(
    (is_target)?"target_grid":"source_grid", (int[2]){3,3}, (int[2]){0,0},
    (double[]){0,1,2}, (double[]){0,1,2}, &grid_id);

  // define points at the vertices of the grid
  int point_id;
  yac_cdef_points_reg2d(
    grid_id, (int[2]){3,3}, YAC_LOCATION_CORNER,
    (double[]){0,1,2}, (double[]){0,1,2}, &point_id);

  // define fields
  int field_ids[3];
  const char * fieldName[3] =
    {"time_op_accu_field", "time_op_avg_field", "time_op_none_field"};
  for (int i = 0; i < 3; ++i )
    yac_cdef_field(
      fieldName[i], comp_id, &point_id, 1, &field_ids[i]);

  int ierror;
  yac_csearch ( &ierror );

  double send_field[3][1][9] =
    {{{ 1, 2, 3, 4, 5, 6, 7, 8, 9}},
     {{10,11,12,13,14,15,16,17,18}},
     {{19,20,21,22,23,24,25,26,27}}};
  double ref_recv_field[3][3][9];
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      for (int k = 0; k < 9; ++k)
        ref_recv_field[i][j][k] = 0.0;

  // do time steps
  for (int t = 0; t < 32; ++t) {

    if (!is_target) {

      for (int i = 0; i < 3; ++i) {

        int info, ierror;
        yac_cput_(field_ids[i], 3, &send_field[0][0][0], &info, &ierror);

        int ref_send_info = (t%4)?0:1;
        if (info != ref_send_info) PUT_ERR("error in yac_cput_: wrong info");
        if (ierror) PUT_ERR("error in yac_cput_: wrong ierror");
      }
    }

    // update ref_recv_field (accu)
    for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 9; ++j)
        ref_recv_field[0][i][j] += send_field[i][0][j];

    // update ref_recv_field (avg)
    double scale = ((t == 0)?1.0:0.25);
    for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 9; ++j)
        ref_recv_field[1][i][j] += send_field[i][0][j] * scale;

    // update ref_recv_field (none)
    for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 9; ++j)
        ref_recv_field[2][i][j] = send_field[i][0][j];

    if (is_target) {

      // target calls get every second timestep
      if ((t % 2) == 0) {
        for (int i = 0; i < 3; ++i) {

          // initialise recv_field
          double recv_field[3][9];
          for (int j = 0; j < 3; ++j)
            for (int k = 0; k < 9; ++k)
              recv_field[j][k] = -1;

          int info, ierror;
          yac_cget_(field_ids[i], 3, &recv_field[0][0], &info, &ierror);

          int ref_recv_info = (t%4)?0:1;
          if (info != ref_recv_info) PUT_ERR("error in yac_cget_: wrong info");
          if (ierror) PUT_ERR("error in yac_cget_: wrong ierror");

          if (info == 1) {
            for (int j = 0; j < 3; ++j)
              for (int k = 0; k < 9; ++k)
                if (fabs(recv_field[j][k] - ref_recv_field[i][j][k]) > 1e-6)
                  PUT_ERR("error in yac_cget_: wrong recv_field");
          } else {
            for (int j = 0; j < 3; ++j)
              for (int k = 0; k < 9; ++k)
                if (recv_field[j][k] != -1)
                  PUT_ERR("error in yac_cget_: wrong recv_field");
          }
        }
      }
    }

    // update send_field
    for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 9; ++j)
        send_field[i][0][j] += 1;

    // clean ref_recv_field at every coupling timestep
    if ((t % 4) == 0)
      for (int i = 0; i < 3; ++i)
        for (int j = 0; j < 3; ++j)
          for (int k = 0; k < 9; ++k)
            ref_recv_field[i][j][k] = 0.0;
  }

  yac_cfinalize ();

  return TEST_EXIT_CODE;
}
